#    This file is part of Eichhoernchen 2.1.
#    Copyright (C) 2020  Carine Dengler
#
#    Eichhoernchen is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
:synopsis: Command-line interpreter.
"""


# standard library imports
import re
import curses
import argparse
import datetime

from io import StringIO
from contextlib import redirect_stderr

# third party imports
# library specific imports
import src.timing

from src import FullName
from src.cutils import mk_menu, mk_stats, readline


def _name(name):
    """Name.

    :param str name: string containing name

    :raises ArgumentTypeError: when string does not contain name

    :returns: name
    :rtype: str
    """
    match = re.match(r"(?:\w|\s|[!#+-?])+", name)
    if match:
        return match.group(0).strip()
    raise argparse.ArgumentTypeError(f"'{name}' does not contain name")


def _tags(tags):
    """Tags.

    :param str tags: string containing tags

    :raises ArgumentTypeError: when string does not contain tags

    :returns: tags
    :rtype: set
    """
    matches = re.findall(r"\[((?:\w|\s|[!#+-?])+)\]", tags)
    if matches:
        return set(tag.strip() for tag in matches)
    raise argparse.ArgumentTypeError(f"'{tags}' does not contain any tags")


def _full_name(full_name):
    """Full name.

    :param str full_name: string containing full name

    :raises ArgumentTypeError: when string does not contain full name

    :returns: full name
    :rtype: FullName
    """
    if not full_name:
        return FullName("", frozenset())
    try:
        return FullName(_name(full_name), _tags(full_name))
    except argparse.ArgumentTypeError:
        return FullName(_name(full_name), frozenset())


def _summand(summand):
    """Summand.

    :param str summand: string containing summand

    :raises ArgumentTypeError: when string does not contain summand

    :returns: summand
    :rtype: FullName
    """
    try:
        return _full_name(summand)
    except argparse.ArgumentTypeError:
        return FullName("", _tags(summand))


def _from(from_):
    """From.

    :param str from_: string containing from

    :raises ArgumentTypeError: when string does not contain from

    :returns: from
    :rtype: str
    """
    try:
        return parse_datetime(
            from_, keywords=("all", "year", "month", "week", "yesterday", "today")
        )
    except ValueError:
        raise argparse.ArgumentTypeError(f"'{from_}' does not contain from")


def _to(to):
    """To.

    :param str to: string containing to

    :returns: to
    :rtype: str
    """
    try:
        return parse_datetime(
            to, keywords=("year", "month", "week", "yesterday", "today")
        )
    except ValueError:
        raise argparse.ArgumentTypeError(f"'{to}' does not contain to")


def _start(start):
    """Start.

    :param str start: string containing start

    :raises ArgumentTypeError: when string does not contain start

    :returns: start
    :rtype: str
    """
    try:
        return parse_datetime(start)
    except ValueError:
        raise argparse.ArgumentTypeError(f"'{start}' does not contain start")


def _end(end):
    """End.

    :param str end: string containing end

    :raises ArgumentTypeError: when string does not contain end

    :returns: end
    :rtype: str
    """
    return _start(end)


def parse_datetime(date_string, keywords=tuple()):
    """Parse date string.

    :param str date_string: date string
    :param tuple keywords: keywords

    :raises ValueError: when date string does not match any format

    :returns: date string
    :rtype: str
    """
    if keywords:
        match = re.match(r"|".join(keywords), date_string)
        if match:
            return match.group(0)
    for format_string in ("%Y-%m-%d %H:%M", "%Y-%m-%d", "%H:%M"):
        try:
            datetime.datetime.strptime(date_string, format_string)
        except ValueError:
            continue
        if format_string == "%H:%M":
            now = datetime.datetime.now()
            date_string = f"{now.year}-{now.month}-{now.day} {date_string}"
        return date_string
    else:
        raise ValueError(f"'{date_string}' does not match any format")


class InterpreterError(Exception):
    """Raised when command-line input cannot be parsed."""

    pass


class Interpreter:
    """Command-line interpreter.

    :cvar dict ARGS: ArgumentParser arguments
    :cvar tuple RESERVED: reserved characters
    :ivar dict subcommands: subcommands
    :ivar Timer timer: timer
    """

    ARGS = {
        "full_name": {
            "type": _full_name,
            "help": "name followed by 0 or more tags, e.g. 'foo[bar]'",
            "metavar": "full name",
        },
        "from_": {
            "type": _from,
            "help": "@([YYYY-MM-DD] [hh:mm]|{all,year,month,week,yesterday,today})",
            "metavar": "from",
        },
        "to": {
            "type": _to,
            "help": "@([YYYY-MM-DD] [hh:mm]|{year,month,week,yesterday,today})",
        },
        "start": {"type": _start, "help": "@YYYY-MM-DD [hh:mm]"},
        "end": {"type": _end, "help": "@YYYY-MM-DD [hh:mm]"},
    }
    RESERVED = ("@",)

    def __init__(self, database, aliases):
        """Initialize command-line interpreter.

        :param str database: pathname of the database
        :param dict aliases: aliases
        """
        self.timer = src.timing.Timer(database)
        self.subcommands = {
            "start": {
                "description": "start task",
                "aliases": aliases.get("start", tuple()),
                "func": self.timer.start,
                "formatter": lambda *args, **kwargs: tuple(),
                "args": {"full_name": self.ARGS["full_name"]},
            },
            "stop": {
                "description": "stop task",
                "aliases": aliases.get("stop", tuple()),
                "func": self.stop,
                "formatter": lambda x: (x,) if x[0][0] else tuple(),
                "args": {},
            },
            "add": {
                "description": "add task",
                "aliases": aliases.get("add", tuple()),
                "func": self.add,
                "formatter": lambda x: (x,),
                "args": {
                    "full_name": self.ARGS["full_name"],
                    "start": self.ARGS["start"],
                    "end": self.ARGS["end"],
                },
            },
            "remove": {
                "description": "remove task",
                "aliases": aliases.get("remove", tuple()),
                "func": self.remove,
                "formatter": lambda x: (x,),
                "args": {
                    "full_name": self.ARGS["full_name"],
                    "from_": {
                        **self.ARGS["from_"],
                        **{"nargs": "?", "default": "today"},
                    },
                },
            },
            "list": {
                "description": "list tasks",
                "aliases": aliases.get("list", tuple()),
                "func": self.timer.list_tasks,
                "formatter": lambda x: (
                    src.output_formatter.pprint_task(y, date=False) for y in x
                ),
                "args": {
                    "full_name": {
                        **self.ARGS["full_name"],
                        **{"nargs": "?", "default": FullName("", set())},
                    },
                    "from_": {
                        **self.ARGS["from_"],
                        **{"nargs": "?", "default": "today"},
                    },
                    "to": {**self.ARGS["to"], **{"nargs": "?", "default": "today"}},
                },
            },
            "show_stats": {
                "description": "show statistics",
                "aliases": aliases.get("show_stats", tuple()),
                "func": self.show_stats,
                "formatter": lambda *args, **kwargs: tuple(),
                "args": {
                    "from_": {
                        **self.ARGS["from_"],
                        **{"nargs": "?", "default": "today"},
                    },
                    "to": {**self.ARGS["to"], **{"nargs": "?", "default": "today"}},
                },
            },
            "edit": {
                "description": "edit task",
                "aliases": aliases.get("edit", tuple()),
                "func": self.edit,
                "formatter": lambda x: (x,),
                "args": {
                    "full_name": self.ARGS["full_name"],
                    "from_": {
                        **self.ARGS["from_"],
                        **{"nargs": "?", "default": "today"},
                    },
                    "to": {
                        **self.ARGS["to"],
                        **{"nargs": "?", "default": "today"},
                    },
                },
            },
            "sum": {
                "description": "sum up total time",
                "aliases": aliases.get("sum", tuple()),
                "func": self.timer.sum_total,
                "formatter": lambda x: (
                    src.output_formatter.pprint_sum(FullName(*y), z) for y, z in x
                ),
                "args": {
                    "summand": {
                        "type": _summand,
                        "help": "full name, name or tag(s) to sum up",
                    },
                    "from_": {
                        **self.ARGS["from_"],
                        **{"nargs": "?", "default": "today"},
                    },
                    "to": {
                        **self.ARGS["to"],
                        **{"nargs": "?", "default": "today"},
                    },
                },
            },
            "aliases": {
                "description": "list aliases",
                "aliases": aliases.get("aliases", tuple()),
                "func": lambda *args, **kwargs: self.aliases(aliases),
                "formatter": lambda x: x,
                "args": {},
            },
            "export": {
                "description": "export tasks",
                "aliases": aliases.get("export", tuple()),
                "func": self.timer.export,
                "formatter": lambda x: (((x, curses.color_pair(4)),),),
                "args": {
                    "ext": {"choices": ("csv", "json"), "metavar": "format"},
                    "from_": {
                        **self.ARGS["from_"],
                        **{"nargs": "?", "default": "today"},
                    },
                    "to": {**self.ARGS["to"], **{"nargs": "?", "default": "today"}},
                },
            },
            "help": {
                "description": "show help",
                "aliases": aliases.get("help", tuple()) + ("?",),
                "args": {},
            },
        }
        self._init_parser(aliases)

    def _init_parser(self, aliases):
        """Initialize parser.

        :param dict aliases: aliases
        """
        self._parser = argparse.ArgumentParser(prog="", add_help=False)
        self._init_subparsers(self._parser, aliases)

    def _init_subparsers(self, parser, aliases):
        """Initialize subparsers.

        :param ArgumentParser parser: command-line interpreter
        :param dict aliases: aliases
        """
        subparsers = parser.add_subparsers()
        subcommands = {}
        for prog, subcommand in self.subcommands.items():
            subcommands[prog] = subparsers.add_parser(
                prog,
                description=subcommand["description"],
                add_help=False,
                aliases=subcommand["aliases"],
            )
            if prog == "help":
                continue
            subcommands[prog].set_defaults(
                func=subcommand["func"], formatter=subcommand["formatter"]
            )
            for arg, kwargs in subcommand["args"].items():
                subcommands[prog].add_argument(arg, **kwargs)
        subcommands["help"].add_argument(
            "command",
            nargs="?",
            choices=(
                tuple(subcommands.keys())
                + tuple(x for y in aliases.values() for x in y)
            ),
        )
        subcommands["help"].set_defaults(
            func=lambda *args, **kwargs: kwargs["command"] or "",
            formatter=lambda x: self.help(x, subcommands, aliases),
        )

    def interpret_line(self, line):
        """Interpret line.

        :param str line: line

        :returns:
        :rtype:
        """
        splits = line.split(maxsplit=1)
        if len(splits) > 1:
            splits = [
                splits[0],
                *[
                    split.strip()
                    for split in re.split(r"|".join(self.RESERVED), splits[1])
                ],
            ]
        # FIXME
        if splits[0] == "show_stats":
            splits = [split for split in splits if split]
        try:
            fp = StringIO()
            with redirect_stderr(fp):
                args = self._parser.parse_args(splits)
        except SystemExit:
            fp.seek(0)
            raise InterpreterError("\t".join(line.strip() for line in fp.readlines()))
        return args.formatter(
            args.func(
                **{
                    k: v
                    for k, v in vars(args).items()
                    if k not in ("func", "formatter")
                }
            )
        )

    def _flatten_items(self, items):
        """Flatten menu items.

        :param list items: items

        :returns: flattened items
        :rtype: str
        """
        return "".join(item[0] for item in items)

    def _pick_task(
        self, full_name=FullName("", frozenset()), from_="today", to="today"
    ):
        """Pick task.

        :param FullName full_name: full name
        :param str from_: from
        :param str to: to

        :returns: task and pretty-printed task
        :rtype: tuple
        """
        tasks = list(self.timer.list_tasks(full_name=full_name, from_=from_, to=to))
        if not tasks:
            raise RuntimeError("no task")
        items = list(
            self._flatten_items(src.output_formatter.pprint_task(task))
            for task in tasks
        )
        i = mk_menu(items)
        if i < 0:
            raise RuntimeError("aborted removing task")
        return tasks[i], items[i]

    def add(self, full_name=FullName("", frozenset()), start="", end=""):
        """Add task.

        :param FullName full_name: full name
        :param str from_: from
        :param str to: to

        :returns: confirmation or error message
        :rtype: tuple
        """
        try:
            task = self.timer.add(full_name=full_name, start=start, end=end)
            return src.output_formatter.pprint_task(task)
        except ValueError as exception:
            return ((str(exception), curses.color_pair(5)),)

    def remove(self, full_name=FullName("", frozenset()), from_="today", to="today"):
        """Choose task to remove.

        :param FullName full_name: full name
        :param str from_: from
        :param str to: to

        :returns: confirmation or error message
        :rtype: tuple
        """
        try:
            task, item = self._pick_task(full_name=full_name, from_=from_, to=to)
        except RuntimeError as exception:
            return ((str(exception), curses.color_pair(5)),)
        self.timer.remove(task)
        return ((f"removed {item}", curses.color_pair(4)),)

    def edit(self, full_name=FullName("", frozenset()), from_="today", to="today"):
        """Choose task to edit.

        :param FullName full_name: full name
        :param str from_: from
        :param str to: to

        :returns: confirmation or error message
        :rtype: tuple
        """
        try:
            task, item = self._pick_task(full_name=full_name, from_=from_, to=to)
        except RuntimeError as exception:
            return ((str(exception), curses.color_pair(5)),)
        if not task:
            return (("no task", curses.color_pair(0)),)
        actions = ("name", "tags", "start", "end")
        i = mk_menu(actions)
        if i < 0:
            return (("aborted editing task", curses.color_pair(5)),)
        arg = (_name, _tags, _from, _to)[i](readline(prompt=f"new {actions[i]} >"))
        if actions[i] in ("start", "end"):
            arg = datetime.datetime.strptime(arg, "%Y-%m-%d %H:%M")
        return src.output_formatter.pprint_task(self.timer.edit(task, actions[i], arg))

    def stop(self):
        """Stop task.

        :returns: confirmation or error message
        :rtype: tuple
        """
        if self.timer.task.name:
            self.timer.stop()
            return (("", curses.color_pair(0)),)
        return (("no running task", curses.color_pair(0)),)

    def help(self, subcommand, subcommands, aliases):
        """Show help message.

        :param str subcommand: subcommand
        :param dict subcommands: subcommands
        :param dict aliases: aliases

        :returns: help message
        :rtype: tuple
        """
        if subcommand:
            if subcommand not in subcommands:
                for k, v in aliases.items():
                    if subcommand in v:
                        subparser = subcommands[k]
                        break
            else:
                subparser = subcommands[subcommand]
            return (
                ((help, curses.color_pair(4)),)
                for help in subparser.format_help().split("\n")
            )
        return (
            ((usage.strip(), curses.color_pair(4)),)
            for usage in self._parser.format_usage().split("\n")
            if usage
        )

    def aliases(self, aliases):
        """List aliases.

        :param dict aliases: aliases

        :returns: aliases
        :rtype: list
        """
        return (
            (("alias\tcommand", curses.color_pair(4)),),
            tuple(),
            *(
                ((f"{alias}\t{k}", curses.color_pair(4)),)
                for k, v in aliases.items()
                for alias in v
            ),
        )

    def convert_to_date_string(self, endpoint):
        """Convert to date string.

        :param str endpoint: endpoint

        :returns: date string
        :rtype: str
        """
        today = datetime.datetime.today()
        if endpoint == "today":
            return today.strftime("%Y-%m-%d")
        if endpoint == "yesterday":
            return (today - datetime.timedelta(days=1)).strftime("%Y-%m-%d")
        if endpoint == "week":
            return (today - datetime.timedelta(days=today.weekday())).strftime(
                "%Y-%m-%d"
            )
        if endpoint == "month":
            return f"{today.year}-{today.month:02}-01"
        if endpoint == "year":
            return f"{today.year}-01-01"

    def pprint_heading(self, from_, to):
        """Pretty-print heading.

        :param str from_: from
        :param str to: to

        :returns: heading
        :rtype: str
        """
        if from_ == "all":
            from_ = (
                sorted(
                    self.timer.list_tasks(from_=from_), key=lambda x: x.time_span[0]
                )[0]
                .time_span[0]
                .strftime("%Y-%m-%d")
            )
        ranges = ("year", "month", "week", "yesterday", "today")
        if from_ in ranges:
            from_ = self.convert_to_date_string(from_)
        if to in ranges:
            to = self.convert_to_date_string(to)
        from_ = datetime.datetime.strptime(from_, "%Y-%m-%d").strftime("%a %d %b %Y")
        to = datetime.datetime.strptime(to, "%Y-%m-%d").strftime("%a %d %b %Y")
        if from_ != to:
            return ((f"overview {from_} - {to}".upper(), curses.color_pair(0)),)
        return ((f"overview {from_}".upper(), curses.color_pair(0)),)

    def show_stats(self, from_="today", to="today"):
        """Show statistics.

        :param str from_: from
        :param str to: to
        """
        heading = self.pprint_heading(from_, to)
        stats = (
            heading,
            ((f"{'—'*len(heading[0][0])}", curses.color_pair(0)),),
            (("", curses.color_pair(0)),),
        )
        tasks = tuple(self.timer.list_tasks(from_=from_, to=to))
        stats += (
            ((f"{len(tasks)} task(s)", curses.color_pair(0)),),
            *(
                src.output_formatter.pprint_task(task, date=(from_ != to))
                for task in tasks
            ),
        )
        tags = {tuple(task.tags) for task in tasks if task.tags}
        pprinted_tags = tuple(
            src.output_formatter.pprint_tags(tuple_)
            for tuple_ in sorted(tags, key=lambda x: len(x))
        )
        stats += (
            ((f"{len(pprinted_tags)} tag(s)", curses.color_pair(0)),),
            *pprinted_tags,
            (("", curses.color_pair(0)),),
        )
        tasks = {(task.name, tuple(task.tags)) for task in tasks}
        sums = sorted(
            (
                sum_
                for name, tags in tasks
                for sum_ in self.timer.sum_total(
                    summand=FullName(name, set(tags)), from_=from_, to=to
                )
            ),
            key=lambda x: x[1],
            reverse=True,
        )
        stats += (
            (("Total runtime task(s)", curses.color_pair(0)),),
            *(
                src.output_formatter.pprint_sum(FullName(*full_name), total)
                for full_name, total in sums
            ),
        )
        sums = sorted(
            (
                sum_
                for tuple_ in tags
                for sum_ in self.timer.sum_total(
                    summand=FullName("", frozenset(tuple_)), from_=from_, to=to
                )
            ),
            key=lambda x: x[1],
            reverse=True,
        )
        stats += (
            (("Total runtime tag(s)", curses.color_pair(0)),),
            *(
                src.output_formatter.pprint_sum(FullName(*full_name), total)
                for full_name, total in sums
            ),
        )
        mk_stats(list(stats))



#    This file is part of Eichhörnchen 2.1.
#    Copyright (C) 2019  Carine Dengler
#
#    Eichhörnchen is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
:synopsis: Basic objects.
"""


# standard library imports
import collections

# third party imports
# library specific imports


# full name composed of name and one or more tags, e.g. foo[bar][baz]
FullName = collections.namedtuple(
    "FullName", ("name", "tags"), defaults=("", frozenset())
)


_Task = collections.namedtuple("Task", ["name", "tags", "time_span"])


class Task(_Task):
    """Task composed of name, one or more tags and its start and end time."""

    __slots__ = ()

    @property
    def total(self):
        """Run time (in seconds).

        :returns: run time (in seconds)
        :rtype: int
        """
        delta = self.time_span[1] - self.time_span[0]
        if delta.days >= 1:
            return delta.seconds + delta.days * (24 * 60 * 60)
        return delta.seconds



#    This file is part of Eichhörnchen 2.1.
#    Copyright (C) 2019  Carine Dengler
#
#    Eichhörnchen is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
:synopsis: Templates.
"""


# standard library imports
import curses

# third party imports
# library specific imports


class Template:
    """Output formatting template."""

    @property
    def name(self):
        return ("{name}", curses.color_pair(1))

    @property
    def tag(self):
        return ("[{tag}]", curses.color_pair(2))

    @property
    def time_span(self):
        return ("{start}-{end}", curses.color_pair(3))

    @property
    def total(self):
        return ("({hours}h{minutes}m)", curses.color_pair(4))

    @property
    def running(self):
        return ("({start}-)", curses.color_pair(3))

    @property
    def prompt(self):
        return ("~>", curses.color_pair(0))



#    This file is part of Eichhörnchen 2.1.
#    Copyright (C) 2020  Carine Dengler
#
#    Eichhörnchen is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
:synopsis: Parser.
"""


# standard library imports
import re
import datetime

# third party imports
# library specific imports


def parse_time(string, time_periods=tuple()):
    """Parse string containing time.

    :param str string: string
    :param tuple time_periods: time periods

    :raises ValueError: when date string does not match any format

    :returns: time
    :rtype: str
    """
    if time_periods:
        match = re.match(r"|".join(time_periods), string)
        if match:
            return match.group(0)
    for format_string in ("%Y-%m-%d %H:%M", "%Y-%m-%d", "%H:%M"):
        try:
            datetime.datetime.strptime(string, format_string)
        except ValueError:
            continue
        if format_string == "%H:%M":
            now = datetime.datetime.now()
            string = f"{now.year}-{now.month}-{now.day} {string}"
        return string
    else:
        raise ValueError(f"string '{string}' does not match any format")


def parse_from(string):
    """Parse string containing start of time span.

    :param str string: string

    :returns: start of time span
    :rtype: str
    """
    return parse_time(
        string, time_periods=("all", "year", "month", "week", "yesterday", "today")
    )


def parse_to(string):
    """Parse string containing end of time span.

    :param str string: string

    :returns: end of time span
    :rtype: str
    """
    return parse_time(
        string, time_periods=("year", "month", "week", "yesterday", "today")
    )


def parse_tags(string):
    """Parse string containing tags.

    :param str string: string

    :raises ValueError: when string does not contain any tags

    :returns: list of tags
    :rtype: set
    """
    matches = re.findall(r"\[((?:\w|\s|[!#+-?])+)\]", string)
    if matches:
        return set(tag.strip() for tag in matches)
    raise ValueError(f"string '{string}' does not contain any tags")


def parse_name(string):
    """Parse string containing name.

    :param str string: string

    :raises ValueError: when string is not name

    :returns: name
    :rtype: str
    """
    match = re.match(r"(?:\w|\s|[!#+-?])+", string)
    if match:
        return match.group(0).strip()
    raise ValueError(f"string '{string}' is not name")



#    Eichhörnchen 2.1
#    Copyright (C) 2018-2019  Carine Dengler
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
:synopsis: Eichhörnchen 2.0.
"""


# standard library imports
import curses
import argparse

# third party imports

# library specific imports
import src.config
import src.cshell

__version__ = "2.1"


def main():
    """Main function."""
    parser = argparse.ArgumentParser(
        prog="eichhoernchen", description="Lightweight curses-based time tracking tool."
    )
    parser.add_argument("-c", "--config", help="use this configuration file")
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    config = src.config.load_config(path=parser.parse_args().config)
    try:
        curses.wrapper(src.cshell.launch, config)
    except Exception as exception:
        print(exception)


main()



#    This file is part of Eichhörnchen 2.1.
#    Copyright (C) 2019  Carine Dengler
#
#    Eichhörnchen is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
:synopsis: Curses-based shell.
"""


# standard library imports
import json
import time
import os
import os.path
import curses
import curses.panel

# third party imports
# library specific imports
from src.interpreter import Interpreter, InterpreterError

from src.output_formatter import pprint_prompt
from src.cutils import (
    initialize_colour,
    mk_panel,
    ResizeError,
    WindowManager,
)


def _loop(stdscr, config):
    """Loop through user interaction.

    :param window stdscr: initial window object
    :param dict config: configuration

    :raises: SystemExit when Control+C is pressed
    """
    panel = mk_panel(*stdscr.getmaxyx(), 0, 0)
    window = panel.window()
    window_mgr = WindowManager(window, banner=True)
    interpreter = Interpreter(
        os.path.join(config["database"]["path"], config["database"]["dbname"]),
        {k: json.loads(v) for k, v in config["aliases"].items()}
        if "aliases" in config
        else {},
    )
    while True:
        try:
            y, _ = window.getyx()
            window_mgr.writeline(
                y, 0, pprint_prompt(task=interpreter.timer.task), move=False
            )
            try:
                line = window_mgr.readline(scroll=True)
            except ResizeError:
                window_mgr.reinitialize()
                continue
            if not line:
                window_mgr.mv_down_or_scroll_down()
                continue
            try:
                output = interpreter.interpret_line(line)
            except (InterpreterError, Warning) as exception:
                window_mgr.mv_down_or_scroll_down()
                window_mgr.writeline(
                    *window_mgr.window.getyx(),
                    ((str(exception), curses.color_pair(5)),)
                )
            else:
                window_mgr.mv_down_or_scroll_down()
                window_mgr.writelines(*window_mgr.window.getyx(), output)
        except KeyboardInterrupt:
            window_mgr.mv_down_or_scroll_down()
            window_mgr.writeline(
                *window_mgr.window.getyx(), (("Goodbye!", curses.color_pair(0)),)
            )
            window.refresh()
            time.sleep(0.5)
            raise SystemExit


def launch(stdscr, config):
    """Launch shell.

    :param window stdscr: initial window object
    :param dict config: configuration
    """
    curses.start_color()
    curses.raw()
    curses.use_default_colors()
    initialize_colour()
    _loop(stdscr, config)



#    This file is part of Eichhörnchen 2.1.
#    Copyright (C) 2018  Carine Dengler
#
#    Eichhörnchen is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
:synopsis: Eichhörnchen SQLite3 database interface.
"""


# standard library imports
import sqlite3
import collections

# third party imports
# library specific imports

TIME_SPAN_COLUMN_DEF = (
    "time_span",
    ("start TIMESTAMP PRIMARY KEY", "end TIMESTAMP CHECK (end > start)"),
)
TAGGED_COLUMN_DEF = (
    "tagged",
    ("tag TEXT", "start TIMESTAMP", "FOREIGN KEY(start) REFERENCES time_span(start)"),
)
RUNNING_COLUMN_DEF = (
    "running",
    ("name TEXT", "start TIMESTAMP", "FOREIGN KEY(start) REFERENCES time_span(start)"),
)
COLUMN_DEF = collections.OrderedDict(
    [TIME_SPAN_COLUMN_DEF, TAGGED_COLUMN_DEF, RUNNING_COLUMN_DEF]
)


def create_table(connection, table, close=True):
    """Create table.

    :param Connection connection: SQLite3 database connection
    :param str table: table
    :param bool close: toggle closing database connection on/off

    :raises ValueError: when table is not defined
    """
    try:
        sql = f"CREATE TABLE IF NOT EXISTS {table} " f"({','.join(COLUMN_DEF[table])})"
        connection.execute(sql)
        connection.commit()
    except sqlite3.Error as exception:
        raise SQLiteError("failed to create table", sql=sql) from exception
    except KeyError:
        raise ValueError(f"'{table}' is not defined")
    finally:
        if close:
            connection.close()


class SQLiteError(Exception):
    """Raised when SQL statement execution fails.

    :ivar str sql: SQL statement
    """

    def __init__(self, *args, sql="", **kwargs):
        """Initialize SQLiteError.

        :param str sql: SQL statement
        """
        super().__init__(*args, **kwargs)
        self.sql = sql


class SQLite:
    """SQLite database interface.

    :ivar str database: SQLite3 database name
    """

    def __init__(self, database):
        """Initialize SQLite3 database interface.

        :param str database: path to SQLite3 database
        """
        self.database = database

    def connect(self):
        """Connect to SQLite3 database.

        :raises SQLiteError: when connection to SQLite3 database fails

        :returns: connection
        :rtype: Connection
        """
        try:
            return sqlite3.connect(
                self.database,
                detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES,
            )
        except sqlite3.Error as exception:
            raise SQLiteError("failed to connect to SQLite3 database") from exception

    def create_database(self):
        """Create SQLite3 database."""
        connection = self.connect()
        for table in COLUMN_DEF:
            create_table(connection, table, close=False)
        connection.close()

    def execute(self, statement, *parameters):
        """Execute SQLite3 statement(s).

        :param str statement: SQLite3 statement
        :param tuple parameters: parameters

        :raises SQLiteError: when SQLite3 statement execution fails

        :returns: rows
        :rtype: list
        """
        try:
            connection = self.connect()
            if len(parameters) > 1:
                rows = connection.executemany(statement, parameters).fetchall()
            else:
                rows = connection.execute(statement, *parameters).fetchall()
        except sqlite3.Error as exception:
            raise SQLiteError(
                "failed to execute SQLite3 statement", sql=statement
            ) from exception
        connection.commit()
        return rows



#    This file is part of Eichhörnchen 2.1.
#    Copyright (C) 2020  Carine Dengler
#
#    Eichhörnchen is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
:synopsis: Curses utils.
"""


# standard library imports
import os
import unicodedata
import curses
import curses.panel

# third party imports
# library specific imports


BANNER = "Welcome to Eichhörnchen.\tType help or ? to list commands."


class ResizeError(Exception):
    """Raised when window has been resized."""

    pass


class WindowManager:
    """Window manager.

    :ivar window window: window
    :ivar list upper_stack: stack (upper window)
    :ivar list lower_stack: stack (lower window)
    :ivar bool box: whether a border is drawn around the edges
    """

    def __init__(self, window, box=False, banner=False):
        """Initialize window manager.

        :param window window: window
        :param bool box: toggle drawing a border around the edges on/off
        :param bool banner: toggle adding a banner on/off
        """
        self.window = window
        self.box = box
        self.banner = banner
        self.reinitialize()
        self.window.idlok(True)
        self.window.keypad(True)
        self.window.scrollok(True)

    def reinitialize(self):
        """Reinitialize window."""
        self.window.clear()
        self.upper_stack = []
        self.lower_stack = []
        y, x = 0, 0
        if self.box:
            self.window.box()
            y += 1
        if self.banner:
            self.window.addstr(y, x, BANNER)
            y, _ = self.window.getyx()
            self.window.move(y + 2, 0)

    def readline(self, y=-1, prompt="", scroll=False, clear=False):
        """Read line.

        :param int y: y position
        :param str prompt: prompt
        :param bool scroll: toggle scrolling on/off
        :param bool clear: toggle clearing line on/off

        :returns: line
        :rtype: str
        """
        _, x = self.window.getyx()
        if y < 0:
            y, _ = self.window.getyx()
        self.window.move(y, x)
        if self.box:
            x += 1
        min_x = x
        if prompt:
            self.window.addstr(y, x, prompt)
            min_x += len(prompt)
        buffer = []
        i = 0
        while True:
            if self.box:
                self.window.box()
            max_y, max_x = self.window.getmaxyx()
            y, x = self.window.getyx()
            ch = self.window.get_wch()
            if ch == curses.KEY_RESIZE:
                raise ResizeError
            if ch == "\x03":
                raise KeyboardInterrupt
            if ch in (curses.KEY_DOWN, curses.KEY_UP):
                if not scroll:
                    continue
            if ch == curses.KEY_DOWN:
                if y < max_y and self.lower_stack:
                    self.scroll_down()
                    if not self.lower_stack:
                        self.window.move(y, len(self.scrapeline(y)[0][0].strip()))
                continue
            if ch == curses.KEY_UP:
                if y > 2 or len(self.upper_stack) > 1:
                    self.scroll_up()
                continue
            if x == 0:
                continue
            if ch in (curses.KEY_ENTER, os.linesep):
                break
            if ch == curses.KEY_LEFT:
                if x > min_x:
                    i -= 1
                    self.window.move(y, x - 1)
                    continue
                if len(buffer) > 0 and i > 0:
                    i -= 1
                    self.window.insstr(y, x, buffer[i])
                    continue
            if ch == curses.KEY_RIGHT:
                if x < max_x - 1:
                    if i < len(buffer):
                        i += 1
                        self.window.move(y, x + 1)
                    continue
                if i < len(buffer) - 1:
                    self.window.delch(y, min_x)
                    i += 1
                    self.window.insstr(y, x, buffer[i])
                    continue
                if self.window.instr(y, x, 1).decode(encoding="utf-8") == buffer[-1]:
                    self.window.delch(y, min_x)
                    i += 1
                    self.window.delch(y, x)
                    continue
            if ch in (curses.KEY_BACKSPACE, 8, 127):
                if x > min_x:
                    i -= 1
                    x -= 1
                    buffer.pop(i)
                    if self.box:
                        self.window.delch(y, max_x - 1)
                    self.window.delch(y, x)
                    if len(buffer) > max_x - (min_x + 2) and i >= x - 1:
                        self.window.insstr(y, min_x, buffer[(i - x) + 1])
                        self.window.move(y, x + 1)
                continue
            if isinstance(ch, int):
                continue
            if unicodedata.category(ch).startswith("C"):
                continue
            buffer.insert(i, ch)
            i += 1
            if x < max_x - 1:
                x += 1
            else:
                self.window.delch(y, min_x)
                self.window.move(y, x - 1)
            self.window.insstr(ch)
            self.window.move(y, x)
        if clear:
            self.window.move(y, 1)
            self.window.clrtoeol()
        if self.box:
            self.window.box()
        return "".join(buffer).strip()

    def writeline(self, y, x, multi_part_line, move=True):
        """Write multi-part line.

        :param int y: y position
        :param int x: x position
        :param tuple multi_part_line: multi-part line
        :param bool move: toggle moving cursor down on/off
        """
        for line, attr in multi_part_line:
            self.window.addstr(y, x, line, attr)
            x += len(line)
        if move:
            self.mv_down_or_scroll_down()

    def writelines(self, y, x, multi_part_lines, move=True):
        """Write multi_part lines.

        :param int y: y position
        :param int x: x position
        :param tuple multi_part_lines: multi-part lines
        :param bool move: toggle moving cursor down on/off
        """
        for multi_part_line in multi_part_lines:
            self.writeline(y, x, multi_part_line, move=move)
            y, _ = self.window.getyx()

    def scrapeline(self, y):
        """Scrape multi-part line.

        :param int y: y position

        :returns: multi-part line
        :rtype: tuple
        """
        _, max_x = self.window.getmaxyx()
        buffer = []
        multi_part_line = []
        current_attr = self.window.inch(y, 0) & curses.A_COLOR
        for x in range(max_x):
            ch = self.window.inch(y, x)
            attr = ch & curses.A_COLOR
            ch = chr(ch & curses.A_CHARTEXT)
            if current_attr != attr:
                multi_part_line.append(("".join(buffer), current_attr))
                buffer = []
                current_attr = attr
            buffer.append(ch)
        buffer = buffer[:-1]
        if buffer:
            multi_part_line.append(("".join(buffer), attr))
        return multi_part_line

    def scroll_up(self):
        """Scroll up."""
        if not self.upper_stack:
            # top of the window
            return
        max_y, _ = self.window.getmaxyx()
        if self.box:
            min_y = 1
            min_x = 1
            max_y -= 1
        else:
            min_y = 0
            min_x = 0
        self.lower_stack.append(self.scrapeline(max_y - 1))
        self.window.scroll(-1)
        self.writeline(min_y, 0, self.upper_stack.pop())
        self.window.move(max_y - 1, min_x)

    def scroll_down(self):
        """Scroll down."""
        max_y, _ = self.window.getmaxyx()
        if self.box:
            min_y = 1
            min_x = 1
            max_y -= 1
        else:
            min_y = 0
            min_x = 0
        self.upper_stack.append(self.scrapeline(min_y))
        self.window.scroll(1)
        if self.box:
            self.window.move(max_y - 1, 0)
            self.window.deleteln()
        self.window.move(max_y - 1, min_x)
        if not self.lower_stack:
            # bottom of the window
            return
        self.writeline(max_y - 1, 0, self.lower_stack.pop(), move=False)
        self.window.move(max_y - 1, min_x)

    def mv_down_or_scroll_down(self):
        """Move cursor down or scroll down."""
        y, _ = self.window.getyx()
        max_y, _ = self.window.getmaxyx()
        if self.box:
            max_y -= 1
        if y < max_y - 1:
            self.window.move(y + 1, 0)
        else:
            self.scroll_down()


def mk_menu(items):
    """Make menu.

    :param list items: items

    :returns: item
    :rtype: int
    """
    if len(items) == 1:
        return 0
    while True:
        panel = mk_panel(*get_menu_dims(*curses.panel.top_panel().window().getmaxyx()))
        window = panel.window()
        window_mgr = WindowManager(window, box=True)
        y, x = window.getyx()
        multi_part_lines = (
            ((f"Pick choice 1...{len(items)}.", curses.color_pair(0)),),
            *tuple(
                ((f"{i}: {item}", curses.color_pair(0)),)
                for i, item in enumerate(items, start=1)
            ),
        )
        window_mgr.writelines(y + 1, x + 1, multi_part_lines)
        y, _ = window_mgr.window.getyx()
        line = ""
        try:
            while line not in (str(i) for i in range(1, len(items) + 1)):
                line = window_mgr.readline(y=y, prompt=">", scroll=True, clear=True)
        except KeyboardInterrupt:
            return -1
        except ResizeError:
            panel.bottom()
            window_mgr.reinitialize()
            continue
        else:
            break
        finally:
            del panel
            curses.panel.update_panels()
    return int(line) - 1


def readline(prompt=""):
    """Read line.

    :param str prompt: prompt

    :returns: line
    :rtype: str
    """
    while True:
        try:
            panel = mk_panel(
                *get_menu_dims(*curses.panel.top_panel().window().getmaxyx())
            )
            window = panel.window()
            window_mgr = WindowManager(window, box=True)
            line = window_mgr.readline(prompt=prompt, y=1)
        except ResizeError:
            panel.bottom()
            window.reinitialize()
            continue
        else:
            return line
        finally:
            del panel
            curses.panel.update_panels()


def mk_stats(stats):
    """Make statistics.

    :param tuple stats: statistics
    """
    while True:
        stats += [
            (("", curses.color_pair(0)),),
            (
                (
                    "Enter 'q' or press Ctrl+C to return to main window",
                    curses.color_pair(0),
                ),
            ),
        ]
        panel = mk_panel(*curses.panel.top_panel().window().getmaxyx(), 0, 0)
        window = panel.window()
        window_mgr = WindowManager(window)
        window_mgr.writelines(*window.getyx(), stats)
        y, _ = window_mgr.window.getyx()
        try:
            line = ""
            while line != "q":
                line = window_mgr.readline(y=y, prompt=">", scroll=True, clear=True)
        except KeyboardInterrupt:
            return
        except ResizeError:
            panel.bottom()
            window_mgr.reinitialize()
            continue
        else:
            break
        finally:
            del panel
            curses.panel.update_panels()
    return


def get_menu_dims(max_y, max_x):
    """Get menu window dimensions.

    :param int max_y: maximum y position
    :param int max_x: maximum x position

    :returns: height, width and initial x, y cursor positions
    :rtype: tuple
    """
    nlines = max_y // 2
    ncols = max_x // 2
    return nlines, ncols, (max_y - nlines) // 2, (max_x - ncols) // 2


def mk_panel(nlines, ncols, begin_y, begin_x):
    """Make panel.

    :param int nlines: number of lines
    :param int ncols: number of columns
    :param int begin_y: initial y position
    :param int begin_x: initial x position

    :returns: panel
    :rtype: panel
    """
    panel = curses.panel.new_panel(curses.newwin(nlines, ncols, begin_y, begin_x))
    curses.panel.update_panels()
    curses.doupdate()
    return panel


def initialize_colour():
    """Initialize colour pairs."""
    colour_pairs = (
        (1, 2, -1),  # name
        (2, 8, -1),  # tags
        (3, 5, -1),  # time span
        (4, 11, -1),  # total
        (5, 9, -1),  # error
    )
    for colour_pair in colour_pairs:
        curses.init_pair(*colour_pair)



#    This file is part of Eichhörnchen 2.1.
#    Copyright (C) 2020  Carine Dengler
#
#    Eichhörnchen is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


# standard library imports
from setuptools import setup

# third party imports
# library specific imports


__version__ = "2.1"


with open("README.md") as fp:
    long_description = fp.read()


setup(
    name="eichhoernchen",
    version=__version__,
    description="Lightweight curses-based time tracking tool.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Carine Dengler",
    author_email="eichhoernchen@pascalin.de",
    url="https://github.com/pascalinde/eichhoernchen",
    packages=["src"],
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    entry_points={"console_scripts": ["eichhoernchen = src.__main__:main"]},
)



#    This file is part of Eichhörnchen 2.1.
#    Copyright (C) 2019  Carine Dengler
#
#    Eichhörnchen is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
:synopsis: Configuration file handling.
"""


# standard library imports
import configparser
import os
import os.path

# third party imports
# library specific imports


class ConfigFound(Exception):
    """Raised when configuration file exists."""

    pass


class BadConfig(Exception):
    """Raised when configuration file contains errors."""

    pass


class ConfigNotFound(Exception):
    """Raised when configuration file does not exist."""

    pass


def _create_config():
    """Create configuration file."""
    path = os.path.join(os.environ["HOME"], ".config/eichhoernchen.ini")
    if os.path.exists(path):
        raise ConfigFound(f"configuration file {path} already exists")
    config = configparser.ConfigParser()
    config["database"] = {
        "dbname": "eichhoernchen.db",
        "path": os.path.join(os.environ["HOME"], ".local/share"),
    }
    with open(path, "w") as fp:
        config.write(fp)


def _validate_config(config):
    """Validate configuration file.

    :param ConfigParser config: configuration

    :raises: BadConfig when required section or key(s) are missing
    """
    check = {"database": {"dbname", "path"}}
    for section, required in check.items():
        try:
            missing = ",".join(
                f"'{key}'" for key in required.difference(set(config[section].keys()))
            )
        except KeyError:
            raise BadConfig(f"required section '{section}' is missing")
        if missing:
            raise BadConfig(f"required key(s) {missing} are missing")


def _read_config(path):
    """Read configuration file.

    :param str path: path to configuration file

    :returns: configuration
    :rtype: ConfigParser
    """
    if not os.path.exists(path):
        raise ConfigNotFound(f"configuration file {path} does not exist")
    else:
        config = configparser.ConfigParser()
        config.read(path)
    _validate_config(config)
    return config


def load_config(path=""):
    """Load configuration file.

    :param str path: path to configuration file

    :raises: SystemExit when configuration file contains errors

    :returns: configuration
    :rtype: dict
    """
    if not path:
        try:
            _create_config()
        except ConfigFound:
            pass
        path = os.path.join(os.environ["HOME"], ".config/eichhoernchen.ini")
    try:
        return _read_config(path)
    except BadConfig as exception:
        raise SystemExit(f"configuration file contains errors:\t{exception}")



#    This file is part of Eichhörnchen 2.1.
#    Copyright (C) 2019  Carine Dengler
#
#    Eichhörnchen is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
:synopsis: Output formatter.
"""


# standard library imports
# third party imports
# library specific imports
from src import FullName, Task
from src.template import Template

TEMPLATE = Template()


def pprint_name(name):
    """Pretty-print name.

    :param str name: name

    :returns: pretty-printed name
    :rtype: tuple
    """
    return ((TEMPLATE.name[0].format(name=name), TEMPLATE.name[1]),)


def pprint_tags(tags):
    """Pretty-print tags.

    :param list tags: tags

    :returns: pretty-printed tags
    :rtype: tuple
    """
    return (
        (
            "".join(TEMPLATE.tag[0].format(tag=tag) for tag in sorted(tags) if tag),
            TEMPLATE.tag[1],
        ),
    )


def pprint_full_name(full_name):
    """Pretty-print full name.

    :param FullName full_name: full name

    :returns: pretty-printed full name
    :rtype: tuple
    """
    return pprint_name(full_name.name) + pprint_tags(full_name.tags)


def pprint_time_span(time_span, date=False):
    """Pretty-print time span.

    :param tuple time_span: time span
    :param bool date: toggle displaying date on/off

    :returns: time span
    :rtype: tuple
    """
    if date:
        if (time_span[1] - time_span[0]).days >= 1:
            start = time_span[0].strftime("%Y-%m-%d %H:%M")
        else:
            start = time_span[0].strftime("%H:%M")
        end = time_span[1].strftime("%H:%M %Y-%m-%d")
    else:
        start = time_span[0].strftime("%H:%M")
        end = time_span[1].strftime("%H:%M")
    return (
        (TEMPLATE.time_span[0].format(start=start, end=end), TEMPLATE.time_span[1]),
    )


def pprint_total(total):
    """Pretty-print total runtime.

    :param int total: runtime (in seconds)

    :returns: pretty-printed runtime
    :rtype: tuple
    """
    hours, minutes = divmod(divmod(total, 60)[0], 60)
    return (
        (TEMPLATE.total[0].format(hours=hours, minutes=minutes), TEMPLATE.total[1]),
    )


def pprint_task(task, date=False):
    """Pretty-print task.

    :param Task task: task
    :param bool date: toggle displaying date on/off

    :returns: pretty-printed task
    :rtype: tuple
    """
    return (
        *pprint_time_span(task.time_span, date=date),
        *pprint_total(task.total),
        *pprint_full_name(FullName(task.name, task.tags)),
    )


def pprint_sum(full_name, total):
    """Pretty-print sum of total runtime.

    :param FullName full_name: full name
    :param int total: runtime (in seconds)

    :returns: pretty-printed runtime
    :rtype: tuple
    """
    return (*pprint_full_name(full_name), *pprint_total(total))


def pprint_running(task):
    """Pretty-print currently running task.

    :param Task task: task

    :returns: pretty-printed currently running task
    :rtype: tuple
    """
    start = task.time_span[0].strftime("%H:%M")
    return ((TEMPLATE.running[0].format(start=start), TEMPLATE.running[1]),)


def pprint_prompt(task=Task("", frozenset(), ())):
    """Pretty-print prompt.

    :param Task task: task

    :returns: pretty-printed prompt
    :rtype: tuple
    """
    if task.name:
        return (
            *pprint_full_name(FullName(task.name, task.tags)),
            *pprint_running(task),
            TEMPLATE.prompt,
        )
    return (TEMPLATE.prompt,)



#    This file is part of Eichhörnchen 2.1.
#    Copyright (C) 2019  Carine Dengler
#
#    Eichhörnchen is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
:synopsis: Timer.
"""


# standard library imports
import csv
import json
import datetime
import itertools
import collections

# third party imports
# library specific imports
import src.sqlite

from src import FullName, Task


class Timer:
    """Timer.

    :ivar SQLite sqlite: SQLite database interface
    :ivar Task task: currently running task
    """

    def __init__(self, database):
        """Initialize timer.

        :param str database: path to SQLite3 database
        """
        self.sqlite = src.sqlite.SQLite(database)
        self.sqlite.create_database()
        self._reset_task()

    def _reset_task(self, task=Task("", frozenset(), ())):
        """Reset currently running task.

        :param Task task: new running task
        """
        self.task = task

    def start(self, full_name):
        """Start task.

        :param FullName full_name: full name

        :raises Warning: when there is already a running task
        """
        if self.task.name:
            raise Warning("there is already a running task")
        start = datetime.datetime.now()
        self.sqlite.execute("INSERT INTO time_span (start) VALUES (?)", (start,))
        self.sqlite.execute(
            "INSERT INTO running (name,start) VALUES (?,?)", (full_name.name, start)
        )
        if full_name.tags:
            self.sqlite.execute(
                "INSERT INTO tagged (tag,start) VALUES (?,?)",
                *[(tag, start) for tag in full_name.tags],
            )
        self._reset_task(task=src.Task(*full_name, (start, None)))

    def stop(self):
        """Stop task."""
        if self.task.name:
            self.sqlite.execute(
                "UPDATE time_span SET end = ? WHERE start = ?",
                (datetime.datetime.now(), self.task.time_span[0]),
            )
            self._reset_task()

    @staticmethod
    def _get_date(endpoint):
        """Get date.

        :param str endpoint: endpoint

        :returns: date
        :rtype: str
        """
        if endpoint == "year":
            return "date('now','localtime','start of year')"
        if endpoint == "month":
            return "date('now','localtime','start of month')"
        if endpoint == "week":
            return "date('now','weekday 0','-7 day')"
        if endpoint == "yesterday":
            return "date('now','start of day','-1 day')"
        if endpoint == "today":
            return "date('now','localtime','start of day')"
        return f"date('{endpoint}','localtime','start of day')"

    def _define_time_period(self, from_, to):
        """Define time period.

        :param str from_: start of time period
        :param str to: end of time period

        :returns: time period
        :rtype: str
        """
        sql = f"WHERE date(time_span.start)<={self._get_date(to)}"
        if from_ == "week":
            return f"{sql} AND date(time_span.start)>{self._get_date(from_)}"
        if from_ != "all":
            return f"{sql} AND date(time_span.start)>={self._get_date(from_)}"
        return sql

    def list_tasks(
        self,
        full_name=FullName("", frozenset()),
        from_="today",
        to="today",
        full_match=True,
    ):
        """List tasks.

        :param FullName full_name: full name
        :param str from_: start of time period
        :param str to: end of time period
        :param bool full_match: toggle matching full name on/off

        :returns: task
        :rtype: Task
        """
        rows = self.sqlite.execute(
            f"""SELECT time_span.start,end,name,tag FROM time_span
            JOIN running ON time_span.start=running.start
            LEFT JOIN tagged ON time_span.start=tagged.start
            {self._define_time_period(from_, to)}
            """
        )
        for (start, end, name), rows in itertools.groupby(rows, key=lambda x: x[:3]):
            tags = {row[-1] for row in rows if row[-1]}
            if full_match:
                if full_name.name and (name, tags) != full_name:
                    continue
            else:
                if full_name.name and name != full_name.name:
                    continue
                if full_name.tags and not tags == full_name.tags:
                    continue
            end = end or datetime.datetime.now()
            yield Task(name, tags, (start, end))

    def sum_total(self, summand=FullName("", frozenset()), from_="today", to="today"):
        """Sum total time up.

        :param FullName summand: full name
        :param str from_: start of time period
        :param str to: end of time period

        :returns: summed up time per task
        :rtype: dict_items
        """
        sum_total = collections.defaultdict(int)
        full_match = summand.name and summand.tags
        for task in self.list_tasks(
            full_name=summand, from_=from_, to=to, full_match=full_match
        ):
            total = int((task.time_span[-1] - task.time_span[0]).total_seconds())
            if summand.name and summand.tags:
                sum_total[(task.name, tuple(task.tags))] += total
                continue
            if summand.name:
                sum_total[(task.name, ("",))] += total
            if summand.tags:
                sum_total[("", tuple(task.tags))] += total
        return sum_total.items()

    def edit(self, task, action, *args):
        """Edit task.

        :param Task task: task to edit
        :param str action: action
        :param str args: arguments

        :raises ValueError: when the task cannot be edited

        :returns: edited task
        :rtype: Task
        """
        name = task.name
        start, end = task.time_span
        tags = task.tags
        is_running = self.task == Task(task.name, tags, (start, None))
        if is_running and action in ("start", "end"):
            raise ValueError(f"cannot edit {action} of a running task")
        if action == "name":
            name = args[0]
            self.sqlite.execute(
                "UPDATE running SET name=? WHERE start=?", (name, start)
            )
        if action == "tags":
            tags = set(args[0])
            self.sqlite.execute("DELETE FROM tagged WHERE start=?", (start,))
            self.sqlite.execute(
                "INSERT INTO tagged (tag,start) VALUES (?,?)",
                *[(tag, start) for tag in tags],
            )
        if action == "end":
            if args[0] <= start:
                raise ValueError(f"{end} is before task's start")
            self.sqlite.execute(
                "UPDATE time_span SET end=? WHERE start=?", (args[0], start)
            )
        if action == "start":
            start = args[0]
            if end <= start:
                raise ValueError(f"{start} is after task's end")
            if self.sqlite.execute("SELECT start FROM running WHERE start=?", (start,)):
                raise ValueError(f"task started at {start} already exists")
            self.sqlite.execute(
                "UPDATE time_span SET start=? WHERE start=?", (start, task.time_span[0])
            )
            self.sqlite.execute(
                "UPDATE running SET start=? WHERE start=?", (start, task.time_span[0])
            )
            self.sqlite.execute(
                "UPDATE tagged SET start=? WHERE start=?", (start, task.time_span[0])
            )
        for task in self.list_tasks(
            full_name=FullName(name, tags), from_=start.date(), to=end.date()
        ):
            if task.time_span == (start, end):
                break
            if is_running:
                self._reset_task(task=task)
        return task

    def remove(self, task):
        """Remove task.

        :param Task task: task to remove

        :raises ValueError: when the task cannot be removed
        """
        if self.task == Task(task.name, task.tags, (task.time_span[0], None)):
            raise ValueError("cannot remove a running task")
        self.sqlite.execute("DELETE FROM running WHERE start=?", (task.time_span[0],))
        self.sqlite.execute("DELETE FROM tagged WHERE start=?", (task.time_span[0],))
        self.sqlite.execute("DELETE FROM time_span WHERE start=?", (task.time_span[0],))

    def add(self, full_name=FullName("", set()), start="", end=""):
        """Add task.

        :param FullName full_name: full name
        :param str start: start of time period
        :param str end: end of time period

        :raises ValueError: when the task cannot be added

        :returns: task
        :rtype: Task
        """
        start = datetime.datetime.strptime(start, "%Y-%m-%d %H:%M")
        end = datetime.datetime.strptime(end, "%Y-%m-%d %H:%M")
        if end <= start:
            raise ValueError(f"task's end ({end}) is before its start ({start})")
        if self.sqlite.execute("SELECT start FROM running WHERE start=?", (start,)):
            raise ValueError(f"task started at {start} already exists")
        self.sqlite.execute(
            "INSERT INTO running (start,name) VALUES (?,?)", (start, full_name.name)
        )
        if full_name.tags:
            self.sqlite.execute(
                "INSERT INTO tagged (tag,start) VALUES (?,?)",
                *[(tag, start) for tag in full_name.tags],
            )
        self.sqlite.execute(
            "INSERT INTO time_span (start,end) VALUES (?,?)", (start, end)
        )
        return Task(full_name.name, full_name.tags, (start, end))

    def _get_row(self, task):
        """Get row.

        :param Task task: task

        :returns: row
        :rtype: tuple
        """
        return (
            task.name,
            task.tags[0],
            task.time_span[0].isoformat(timespec="seconds"),
            task.time_span[1].isoformat(timespec="seconds"),
        )

    def _export_csv(self, filename, tasks):
        """Export tasks to CSV file.

        :param str filename: filename
        :param list tasks: list of tasks
        """
        rows = [
            self._get_row(Task(task.name, (tag,), task.time_span))
            for task in tasks
            for tag in task.tags
            if task.tags
        ]
        rows += [
            self._get_row(Task(task.name, ("",), task.time_span))
            for task in tasks
            if not task.tags
        ]
        with open(filename, mode="w") as fp:
            csv.writer(fp).writerows(rows)

    def _export_json(self, filename, tasks):
        """Export tasks to JSON file.

        :param str filename: filename
        :param list tasks: list of tasks
        """
        with open(filename, mode="w") as fp:
            json.dump(
                [
                    (
                        task.name,
                        list(task.tags),
                        task.time_span[0].isoformat(timespec="seconds"),
                        task.time_span[1].isoformat(timespec="seconds"),
                    )
                    for task in tasks
                ],
                fp,
            )

    def export(self, ext="csv", from_="today", to="today"):
        """Export tasks.

        :param str ext: file format
        :param str from_: start of time period
        :param str to: end of time period

        :returns: confirmation message
        :rtype: str
        """
        tasks = self.list_tasks(from_=from_, to=to)
        filename = f"/tmp/{datetime.datetime.now().strftime('%Y%m%d')}.{ext}"
        if ext == "csv":
            self._export_csv(filename, tasks)
        if ext == "json":
            self._export_json(filename, tasks)
        return f"exported tasks from {from_} to {to} to {filename}"
